#!/usr/bin/env python3
#
#  distributions.py
"""
Iterate over installed distributions.

Third-party distributions are installed into Python's ``site-packages`` directory with tools such as pip_.
Distributions must have a ``*.dist-info`` directory (as defined by :pep:`566`) to be discoverable.

.. _pip: https://pypi.org/project/pip/
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Parts of iter_distributions based on https://github.com/takluyver/entrypoints
#  Copyright (c) 2015 Thomas Kluyver and contributors
#  MIT Licensed
#

# stdlib
import abc
import csv
import posixpath
import re
import sys
from contextlib import suppress
from operator import itemgetter
from typing import (
		TYPE_CHECKING,
		Any,
		Callable,
		Dict,
		Iterable,
		Iterator,
		List,
		Optional,
		Tuple,
		Type,
		TypeVar,
		cast
		)

# 3rd party
import handy_archives
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from domdf_python_tools.utils import divide
from packaging.version import Version

# this package
from dist_meta import metadata, wheel
from dist_meta._utils import _canonicalize, _iter_dist_infos, _parse_version, _parse_wheel_filename
from dist_meta.metadata_mapping import MetadataMapping
from dist_meta.record import FileHash, RecordEntry

_tuplegetter = lambda index, doc: property(itemgetter(index), doc=doc)

if not TYPE_CHECKING:
	with suppress(ImportError):
		# 3rd party
		from _collections import _tuplegetter

__all__ = [
		"get_distribution",
		"iter_distributions",
		"DistributionType",
		"Distribution",
		"WheelDistribution",
		"DistributionNotFoundError",
		"_DT",
		]

_DT = TypeVar("_DT", bound="DistributionType")
_D = TypeVar("_D", bound="Distribution")
_WD = TypeVar("_WD", bound="WheelDistribution")


class DistributionType(abc.ABC):
	"""
	Abstract base class for :class:`~.Distribution`-like objects.

	.. versionchanged:: 0.3.0

		Previously was a :py:obj:`~.typing.Union` representing :class:`~.Distribution` and :class:`~.WheelDistribution`.
		Now a common base class for those two classes, and custom classes providing the same API

	This class implements most of the :func:`collections.namedtuple` API.
	Subclasses must implement ``_fields`` (as a tuple of field names)
	and the :class:`tuple` interface (specifically ``__iter__`` and ``__getitem__``).
	"""

	#: The name of the distribution. No normalization is performed.
	name: str

	#: The version number of the distribution.
	version: Version

	#: A tuple of field names for the "namedtuple".
	_fields: Tuple[str, ...]  # actually a ClassVar, but need to support older Pythons

	#: A mapping of field names to default values.
	_field_defaults: Dict[str, Any]

	# These must be implemented by subclasses, but they can't be abstractmethods
	# otherwise mypy gets confused
	__iter__: Callable
	__getitem__: Callable

	def __init_subclass__(cls: Type["DistributionType"], **kwargs):

		if not getattr(cls, "_fields", ()):
			raise ValueError("'_fields' cannot be empty.")

		ns = cls.__dict__
		field_defaults = getattr(cls, "_field_defaults", {})

		for index, name in enumerate(cls._fields):
			if name in ns:
				field_defaults[name] = ns[name]

		if cls._fields[0] != "name":
			raise ValueError("The first item in '_fields' must be 'name'")
		elif cls._fields[1] != "version":
			raise ValueError("The second item in '_fields' must be 'version'")

		for index, name in enumerate(cls._fields):
			doc = sys.intern(f'Alias for field number {index}')
			setattr(cls, name, _tuplegetter(index, doc))

		cls._field_defaults = field_defaults

	@abc.abstractmethod
	def read_file(self, filename: str) -> str:
		"""
		Read a file from the ``*.dist-info`` directory and return its content.

		:param filename:
		"""

		raise NotImplementedError

	@abc.abstractmethod
	def has_file(self, filename: str) -> bool:
		"""
		Returns whether the ``*.dist-info`` directory contains a file named ``filename``.

		:param filename:
		"""

		raise NotImplementedError

	def _asdict(self) -> Dict[str, Any]:
		"""
		Return a new dict which maps field names to their values.
		"""

		return dict(zip(self._fields, self))

	def __getnewargs__(self) -> Tuple:
		"""
		Return self as a plain tuple. Used by copy and pickle.
		"""

		return tuple(self)

	def _replace(self: _DT, **kwargs) -> _DT:
		"""
		Make a new :class:`~.DistributionType` object, of the same type as this one,
		replacing the specified fields with new values.

		:param iterable:
		"""  # noqa: D400

		result = self._make(map(kwargs.pop, self._fields, self))
		if kwargs:
			raise ValueError(f'Got unexpected field names: {list(kwargs)!r}')
		return result

	@classmethod
	def _make(cls: Type[_DT], iterable) -> _DT:
		"""
		Make a new :class:`~.DistributionType` object, of the same type as this one, from a sequence or iterable.

		:param iterable:
		"""

		return cls(*iterable)

	def get_entry_points(self) -> Dict[str, Dict[str, str]]:  # -> EntryPointMap
		"""
		Returns a mapping of entry point groups to entry points.

		Entry points in the group are contained in a dictionary mapping entry point names to objects.

		:class:`dist_meta.entry_points.EntryPoint` objects can be constructed as follows:

		.. code-block:: python

			for name, epstr in distro.get_entry_points().get("console_scripts", {}).items():
				EntryPoint(name, epstr)
		"""

		# this package
		from dist_meta import entry_points

		if self.has_file("entry_points.txt"):
			return entry_points.loads(self.read_file("entry_points.txt"))
		else:
			return {}

	def get_metadata(self) -> MetadataMapping:
		"""
		Returns the content of the ``*.dist-info/METADATA`` file.
		"""

		return metadata.loads(self.read_file("METADATA"))

	def get_wheel(self) -> Optional[MetadataMapping]:
		"""
		Returns the content of the ``*.dist-info/WHEEL`` file, or :py:obj:`None` if the file does not exist.

		The file will only be present if the distribution was installed from a :pep:`wheel <427>`.
		"""  # noqa: RST399

		if self.has_file("WHEEL"):
			return wheel.loads(self.read_file("WHEEL"))
		else:
			return None

	def get_record(self) -> Optional[List[RecordEntry]]:
		"""
		Returns the parsed content of the ``*.dist-info/RECORD`` file, or :py:obj:`None` if the file does not exist.

		:returns: A :class:`dist_meta.record.RecordEntry` object for each line in the record
			(i.e. each file in the distribution).
			This includes files in the ``*.dist-info`` directory.
		"""  # noqa: RST399

		if self.has_file("RECORD"):
			content = self.read_file("RECORD").splitlines()
			output = []

			for line in csv.reader(content):
				name, hash_, size_str, *_ = line
				entry = RecordEntry(
						name.strip(),
						hash=FileHash.from_string(hash_) if hash_ else None,
						size=int(size_str) if size_str else None,
						)
				output.append(entry)

			return output
		else:
			return None

	def __repr__(self):
		"""
		Returns a string representation of the :class:`~.DistributionType`.
		"""

		return f"<{self.__class__.__name__}({self.name!r}, {self.version!r})>"


class Distribution(DistributionType, Tuple[str, Version, PathPlus]):
	"""
	Represents an installed Python distribution.

	:param name: The name of the distribution.
	"""

	#: The name of the distribution. No normalization is performed.
	name: str

	#: The version number of the distribution.
	version: Version

	#: The path to the ``*.dist-info`` directory in the file system.
	path: PathPlus

	__slots__ = ()
	_fields = ("name", "version", "path")

	def __new__(
			cls: Type[_D],
			name: str,
			version: Version,
			path: PathPlus,
			) -> _D:
		"""
		Construct a new :class:`~.Distribution` object.

		:rtype: :class:`~.Distribution`
		"""

		# If this is super().__new__ it breaks on PyPy
		return tuple.__new__(cls, (name, version, path))

	@classmethod
	def from_path(cls: Type[_D], path: PathLike) -> _D:
		"""
		Construct a :class:`~.Distribution` from a filesystem path to the ``*.dist-info`` directory.

		:param path:

		:rtype: :class:`~.Distribution`
		"""

		path = PathPlus(path)
		distro_name_version = path.stem
		name, version = divide(distro_name_version, '-')
		return cls(name, _parse_version(version), path)

	def read_file(self, filename: str) -> str:
		"""
		Read a file from the ``*.dist-info`` directory and return its content.

		:param filename:
		"""

		return (self.path / filename).read_text()

	def has_file(self, filename: str) -> bool:
		"""
		Returns whether the ``*.dist-info`` directory contains a file named ``filename``.

		:param filename:
		"""

		return (self.path / filename).is_file()

	def get_record(self) -> Optional[List[RecordEntry]]:
		"""
		Returns the parsed content of the ``*.dist-info/RECORD`` file, or :py:obj:`None` if the file does not exist.

		:returns: A :class:`dist_meta.record.RecordEntry` object for each line in the record
			(i.e. each file in the distribution).
			This includes files in the ``*.dist-info`` directory.
		"""  # noqa: RST399

		if self.has_file("RECORD"):
			content = self.read_file("RECORD").splitlines()
			output = []

			for line in csv.reader(content):
				name, hash_, size_str, *_ = line
				entry = RecordEntry(
						name.strip(),
						hash=FileHash.from_string(hash_) if hash_ else None,
						size=int(size_str) if size_str else None,
						distro=self,
						)
				output.append(entry)

			return output
		else:
			return None


class WheelDistribution(DistributionType, Tuple[str, Version, PathPlus, handy_archives.ZipFile]):
	"""
	Represents a Python distribution in :pep:`wheel <427>` form.

	:param name: The name of the distribution.


	A :class:`~.WheelDistribution` can be used as a contextmanager,
	which will close the underlying :class:`zipfile.ZipFile` when exiting
	the :keyword:`with` block.
	"""  # noqa: RST399

	#: The name of the distribution. No normalization is performed.
	name: str

	#: The version number of the distribution.
	version: Version

	#: The path to the ``.whl`` file.
	path: PathPlus

	#: The opened zip file.
	wheel_zip: handy_archives.ZipFile

	__slots__ = ()
	_fields = ("name", "version", "path", "wheel_zip")

	def __new__(
			cls: Type[_WD],
			name: str,
			version: Version,
			path: PathPlus,
			wheel_zip: handy_archives.ZipFile,
			) -> _WD:
		"""
		Construct a new :class:`~.WheelDistribution` object.

		:rtype: :class:`~.WheelDistribution`
		"""

		# If this is super().__new__ it breaks on PyPy
		return tuple.__new__(cls, (name, version, path, wheel_zip))

	@classmethod
	def from_path(cls: Type[_WD], path: PathLike, **kwargs) -> _WD:
		r"""
		Construct a :class:`~.WheelDistribution` from a filesystem path to the ``.whl`` file.

		:param path:
		:param \*\*kwargs: Additional keyword arguments passed to :class:`zipfile.ZipFile`.

		:rtype: :class:`~.WheelDistribution`
		"""

		path = PathPlus(path)
		name, version, *_ = _parse_wheel_filename(path)
		wheel_zip = handy_archives.ZipFile(path, 'r', **kwargs)

		return cls(name, version, path, wheel_zip)

	def __enter__(self: _WD) -> _WD:
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.wheel_zip.close()

	def read_file(self, filename: str) -> str:
		"""
		Read a file from the ``*.dist-info`` directory and return its content.

		:param filename:
		"""

		dist_info = f"{self.name}-{self.version}.dist-info"
		try:
			return self.wheel_zip.read_text(posixpath.join(dist_info, filename))
		except FileNotFoundError:
			filenames = _get_case_insensitive_matches(filename, self)
			if not filenames:
				raise

			return self.wheel_zip.read_text(filenames[0])

	def has_file(self, filename: str) -> bool:
		"""
		Returns whether the ``*.dist-info`` directory contains a file named ``filename``.

		:param filename:
		"""

		dist_info = f"{self.name}-{self.version}.dist-info"

		if posixpath.join(dist_info, filename) in self.wheel_zip.namelist():
			return True
		else:
			return bool(_get_case_insensitive_matches(filename, self))

	def get_wheel(self) -> MetadataMapping:
		"""
		Returns the content of the ``*.dist-info/WHEEL`` file.

		:raises FileNotFoundError: if the file does not exist.
		"""

		return wheel.loads(self.read_file("WHEEL"))

	def get_record(self) -> List[RecordEntry]:
		"""
		Returns the parsed content of the ``*.dist-info/RECORD`` file, or :py:obj:`None` if the file does not exist.

		:returns: A :class:`dist_meta.record.RecordEntry` object for each line in the record
			(i.e. each file in the distribution).
			This includes files in the ``*.dist-info`` directory.

		:raises FileNotFoundError: if the file does not exist.
		"""

		content = self.read_file("RECORD").splitlines()
		output = []

		for line in csv.reader(content):
			name, hash_, size_str, *_ = line
			entry = RecordEntry(
					name,
					hash=FileHash.from_string(hash_) if hash_ else None,
					size=int(size_str) if size_str else None,
					)
			output.append(entry)

		return output


def iter_distributions(path: Optional[Iterable[PathLike]] = None) -> Iterator[Distribution]:
	"""
	Returns an iterator over installed distributions on ``path``.

	:param path: The directories entries to search for distributions in.
	:default path: :py:data:`sys.path`
	"""

	if path is None:  # pragma: no cover
		path = sys.path

	# Distributions found earlier in path will shadow those with the same name found later.
	# If these distributions used different module names, it may actually be possible to import both,
	# but in most cases this shadowing will be correct.
	distro_names_seen = set()

	for folder in map(PathPlus, path):
		if not folder.is_dir():
			continue

		for subdir in _iter_dist_infos(folder):
			distro = Distribution.from_path(subdir)

			normalized_name = _canonicalize(distro.name)

			if normalized_name in distro_names_seen:
				continue

			distro_names_seen.add(normalized_name)

			yield distro


def get_distribution(
		name: str,
		path: Optional[Iterable[PathLike]] = None,
		) -> Distribution:
	"""
	Returns a :class:`~.Distribution` instance for the distribution with the given name.

	:param name:
	:param path: The directories entries to search for distributions in.
	:default path: :py:data:`sys.path`

	:rtype:
	"""

	for distro in iter_distributions(path=path):
		if _canonicalize(distro.name) == _canonicalize(name):
			return distro

	raise DistributionNotFoundError(name)


class DistributionNotFoundError(ValueError):
	"""
	Raised when a distribution cannot be located.
	"""


def _get_case_insensitive_matches(filename: str, distribution: DistributionType) -> List[str]:
	"""
	Return a list of matching filenames for a case insentitive search of ``*.dist-info`` directories.

	Some backends, e.g. poetry, allow CamelCase for the wheel filename
	but use lowercase for the ``*.dist-info`` directory

	:param filename:
	:param distribution:
	"""

	cre = re.compile(
			rf"^(?i:{re.escape(distribution.name)})-"
			rf"{re.escape(str(distribution.version))}\.dist-info/{re.escape(filename)}$"
			)
	namelist = cast(WheelDistribution, distribution).wheel_zip.namelist()
	return [fn for fn in namelist if cre.match(fn)]



#!/usr/bin/env python3
#
#  entry_points.py
"""
Parser and emitter for ``entry_points.txt``.

.. note::

	The functions in this module will only parse well-formed ``entry_points.txt`` files,
	and may return unexpected values if passed malformed input.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Parts based on https://github.com/python/importlib_metadata
#  Copyright 2017-2019 Jason R. Coombs, Barry Warsaw
#  Licensed under the Apache License, Version 2.0
#
#  EntryPoint based on https://github.com/takluyver/entrypoints
#  Copyright (c) 2015 Thomas Kluyver and contributors
#  MIT Licensed
#

# stdlib
import functools
import importlib
import itertools
import re
from typing import Dict, Iterable, Iterator, List, Mapping, NamedTuple, Optional, Sequence, Tuple, TypeVar, Union

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.stringlist import StringList
from domdf_python_tools.typing import PathLike
from domdf_python_tools.utils import divide

# this package
from dist_meta._utils import _cache
from dist_meta.distributions import Distribution

__all__ = [
		"lazy_load",
		"lazy_loads",
		"load",
		"loads",
		"dump",
		"dumps",
		"get_entry_points",
		"get_all_entry_points",
		"EntryPoint",
		]

_EP = TypeVar("_EP", bound="EntryPoint")

#: Type hint for a lazily evaluated iterator of entry points.
EntryPointIterator = Iterator[Tuple[str, Iterator[Tuple[str, str]]]]

#: Type hint for a mapping of entry point groups to mappings of entry point names to entry point objects.
EntryPointMap = Dict[str, Dict[str, str]]


class _Section:

	def __init__(self):
		self.section: Optional[str] = None

	def __call__(self, line) -> Optional[str]:
		if line.startswith('[') and line.endswith(']'):
			# new section
			self.section = line.strip("[]")
			return None

		return self.section


def _parse_value(line: str) -> Tuple[str, str]:
	name, obj = divide(line, '=')
	return name.strip(), obj.strip()


def lazy_loads(rawtext: str) -> EntryPointIterator:
	"""
	Parse the entry points from the given text lazily.

	:param rawtext:

	:returns: An iterator over ``(group, entry_point)`` tuples, where ``entry_point``
		is an iterator over ``(name, object)`` tuples.
	"""

	lines = filter(None, map(str.strip, rawtext.splitlines()))

	for section, values in itertools.groupby(lines, _Section()):
		if section is not None:
			yield section, map(_parse_value, values)


def lazy_load(filename: PathLike) -> EntryPointIterator:
	"""
	Parse the entry points from the given file lazily.

	:param filename:

	:returns: An iterator over ``(group, entry_point)`` tuples, where ``entry_point``
		is an iterator over ``(name, object)`` tuples.
	"""

	filename = PathPlus(filename)
	return lazy_loads(filename.read_text())


@_cache
def loads(rawtext: str) -> EntryPointMap:
	"""
	Parse the entry points from the given text.

	:param rawtext:

	:returns: A mapping of entry point groups to entry points.

		Entry points in each group are contained in a dictionary mapping entry point names to objects.

		:class:`dist_meta.entry_points.EntryPoint` objects can be constructed as follows:

		.. code-block:: python

			for name, epstr in distro.get_entry_points().get("console_scripts", {}).items():
				EntryPoint(name, epstr)
	"""

	eps = lazy_loads(rawtext)
	return {k: dict(v) for k, v in eps}


def load(filename: PathLike) -> EntryPointMap:
	"""
	Parse the entry points from the given file.

	:param filename:

	:returns: A mapping of entry point groups to entry points.

		Entry points in each group are contained in a dictionary mapping entry point names to objects.

		:class:`dist_meta.entry_points.EntryPoint` objects can be constructed as follows:

		.. code-block:: python

			for name, epstr in distro.get_entry_points().get("console_scripts", {}).items():
				EntryPoint(name, epstr)
	"""

	filename = PathPlus(filename)
	return loads(filename.read_text())


def dumps(entry_points: Union[EntryPointMap, Dict[str, Sequence["EntryPoint"]]]) -> str:
	"""
	Construct an ``entry_points.txt`` file for the given grouped entry points.

	:param entry_points:  A mapping of entry point groups to entry points.

		Entry points in each group are contained in a dictionary mapping entry point names to objects,
		or in a list of :class:`~.EntryPoint` objects.
	"""

	output = StringList()

	for group in entry_points:
		output.append(f"[{group}]")

		for name in entry_points[group]:
			if isinstance(name, EntryPoint):
				output.append(f"{name.name} = {name.value}")
			else:
				output.append(f"{name} = {entry_points[group][name]}")  # type: ignore

		output.blankline(ensure_single=True)

	return str(output)


def dump(
		entry_points: Union[EntryPointMap, Dict[str, Sequence["EntryPoint"]]],
		filename: PathLike,
		) -> int:
	"""
	Construct an ``entry_points.txt`` file for the given grouped entry points, and write it to ``filename``.

	:param entry_points:  A mapping of entry point groups to entry points.

		Entry points in each group are contained in a dictionary mapping entry point names to objects,
		or in a list of :class:`~.EntryPoint` objects.

	:param filename:
	"""

	filename = PathPlus(filename)
	return filename.write_text(dumps(entry_points))


def get_entry_points(
		group: str,
		path: Optional[Iterable[PathLike]] = None,
		) -> Iterator["EntryPoint"]:
	"""
	Returns an iterator over :class:`entrypoints.EntryPoint` objects in the given group.

	:param group:
	:param path: The directories entries to search for distributions in.
	:default path: :py:data:`sys.path`
	"""

	# this package
	from dist_meta.distributions import iter_distributions

	for distro in iter_distributions(path=path):

		eps = distro.get_entry_points()

		if group in eps:
			for name, epstr in eps[group].items():
				yield EntryPoint(name, epstr, group=group, distro=distro)


def get_all_entry_points(path: Optional[Iterable[PathLike]] = None, ) -> Dict[str, List["EntryPoint"]]:
	"""
	Returns a mapping of entry point groups to entry points for all installed distributions.

	:param path: The directories entries to search for distributions in.
	:default path: :py:data:`sys.path`
	"""

	# this package
	from dist_meta.distributions import iter_distributions

	grouped_eps: Dict[str, List[EntryPoint]] = {}

	for distro in iter_distributions(path=path):

		eps = distro.get_entry_points()

		for group_name in eps:
			group = grouped_eps.setdefault(group_name, [])

			for name, epstr in eps[group_name].items():
				group.append(EntryPoint(name, epstr, group_name, distro))

	return grouped_eps


_entry_point_pattern = re.compile(
		r"""
(?P<modulename>\w+(\.\w+)*)
(:(?P<objectname>\w+(\.\w+)*))?
\s*
(\[(?P<extras>.+)])?
$
""",
		re.VERBOSE
		)


class EntryPoint(NamedTuple):
	"""
	Represents a single entry point.
	"""

	#: The name of the entry point.
	name: str

	#: The value of the entry point, in the form ``module.submodule:attribute``.
	value: str

	#: The group the entry point belongs to.
	group: Optional[str] = None

	#: The distribution the entry point belongs to.
	distro: Optional["Distribution"] = None

	def load(self):
		"""
		Load the object referred to by this entry point.

		If only a module is indicated by the value, return that module.
		Otherwise, return the named object.
		"""

		match = _entry_point_pattern.match(self.value)
		if not match:
			raise ValueError(f"Malformed entry point {self.value!r}")

		module_name, object_name = match.group("modulename", "objectname")
		obj = importlib.import_module(module_name)

		if object_name:
			for attr in object_name.split('.'):
				obj = getattr(obj, attr)

		return obj

	@property
	def extras(self) -> List[str]:
		"""
		Returns the list of extras associated with the entry point.
		"""

		match = _entry_point_pattern.match(self.value)
		if not match:
			raise ValueError(f"Malformed entry point {self.value!r}")

		extras = match.group("extras")

		if extras is not None:
			return re.split(r',\s*', extras)

		return []

	@property
	def module(self) -> str:
		"""
		The module component of :class:`self.value <.EntryPoint>`.
		"""

		# TODO: proper xref

		match = _entry_point_pattern.match(self.value)
		if not match:
			raise ValueError(f"Malformed entry point {self.value!r}")

		return match.group("modulename")

	@property
	def attr(self) -> str:
		"""
		The object/attribute component of :class:`self.value <.EntryPoint>`.

		:rtype:

		.. latex:clearpage::
		"""

		# TODO: proper xref

		match = _entry_point_pattern.match(self.value)
		if not match:
			raise ValueError(f"Malformed entry point {self.value!r}")

		return match.group("objectname")

	@classmethod
	def from_mapping(
			cls,
			mapping: Mapping[str, str],
			*,
			group: Optional[str] = None,
			distro: Optional["Distribution"] = None,
			) -> List["EntryPoint"]:
		"""
		Returns a list of :class:`~.EntryPoint` objects constructed from values in ``mapping``.

		:param mapping: A mapping of entry point names to values,
			where values are in the form ``module.submodule:attribute``.
		:param group: The group the entry points belong to.
		:param distro: The distribution the entry points belong to.
		"""

		output = []

		for name, value in mapping.items():
			output.append(EntryPoint(name, value, group, distro))

		return output



#!/usr/bin/env python3
#
#  metadata.py
"""
Parse and create ``*dist-info/METADATA`` files.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import inspect
import re
import sys
import textwrap
from typing import List

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from domdf_python_tools.utils import divide

# this package
from dist_meta.metadata_mapping import MetadataEmitter, MetadataMapping

__all__ = ["dump", "dumps", "load", "loads", "MissingFieldError"]

WSP = " \t"
DELIMITER = "\n\n"
NEWLINE_MARK = '\uf8ff'

_unfold_re = re.compile(rf"\n([{WSP}])")


def _clean_desc(lines: List[str], wsp: str) -> List[str]:
	#  Adapted from inspect.cleandoc
	#  Licensed under the Python Software Foundation License Version 2.
	#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
	#  Copyright © 2000 BeOpen.com. All rights reserved.
	#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
	#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.

	assert len(wsp) == 1

	# Find minimum indentation of any non-blank lines after first line.
	margin = sys.maxsize

	include_first_line = False

	for line in lines[1:]:
		content = len(line.lstrip(wsp))
		if content:
			indent = len(line) - content
			margin = min(margin, indent)

	# Remove indentation.
	if margin < sys.maxsize:
		if lines[0][:margin] == (wsp * margin):
			include_first_line = True

		for i in range(not include_first_line, len(lines)):
			lines[i] = lines[i][margin:]

	return lines


class MissingFieldError(ValueError):
	"""
	Raised when a required field is missing.
	"""


def loads(rawtext: str) -> MetadataMapping:
	"""
	Parse Python core metadata from the given string.

	:param rawtext:

	:returns: A mapping of the metadata fields, and the long description
	"""

	if DELIMITER in rawtext:
		rawtext, body = rawtext.split(DELIMITER, maxsplit=1)
	else:
		body = ''

	# unfold per RFC 5322 § 2.2.3
	rawtext = _unfold_re.sub(fr"{NEWLINE_MARK}\1", rawtext)

	file_content: List[str] = rawtext.split('\n')
	file_content.reverse()

	fields: MetadataMapping = MetadataMapping()

	while file_content:
		line = file_content.pop()
		if not line:
			continue

		field_name, field_value = divide(line, ':')

		if field_name.lower() != "description":
			fields[field_name] = field_value.replace(NEWLINE_MARK, '').lstrip()
		else:
			# Unwrap
			description_lines = field_value.split(NEWLINE_MARK)
			description_lines = _clean_desc(description_lines, ' ')
			description_lines = _clean_desc(description_lines, '\t')
			description_lines = _clean_desc(description_lines, '|')

			# Remove any trailing or leading blank lines.
			while description_lines and not description_lines[-1]:
				description_lines.pop()
			while description_lines and not description_lines[0]:
				description_lines.pop(0)

			field_value = '\n'.join(description_lines).strip() + '\n'

			fields["Description"] = field_value

	if body.strip():
		if "Description" in fields:
			raise ValueError(
					"A value was given for the 'Description' field "
					"but the body of the file is not empty."
					)
		else:
			fields["Description"] = body.strip() + '\n'

	for required_field in ["Metadata-Version", "Name", "Version"]:
		if required_field not in fields:
			raise MissingFieldError(f"No {required_field!r} field was provided.")

	return fields


def load(filename: PathLike) -> MetadataMapping:
	"""
	Parse Python core metadata from the given file.

	:param filename:

	:returns: A mapping of the metadata fields, and the long description
	"""

	filename = PathPlus(filename)
	return loads(filename.read_text())


def dumps(fields: MetadataMapping) -> str:
	"""
	Construct Python core metadata from the given fields.

	:param fields:
	"""

	output = MetadataEmitter(fields)

	if "Metadata-Version" in fields:
		version = float(fields["Metadata-Version"])
		output.append(f"Metadata-Version: {fields['Metadata-Version']}")
	else:
		raise MissingFieldError("No 'Metadata-Version' field was provided.")

	if version < 2.1:
		raise ValueError("'dump_metadata' only supports metadata version 2.1 and above.")

	for required_field in ["Name", "Version"]:
		if required_field in fields:
			output.append(f"{required_field}: {fields[required_field]}")
		else:
			raise MissingFieldError(f"No {required_field!r} field was provided.")

	if version >= 2.2:
		output.add_multiple("Dynamic")

	# General Meta
	output.add_single("Summary")
	output.add_single("Author")
	output.add_single("Author-email")
	output.add_single("Maintainer")
	output.add_single("Maintainer-email")
	output.add_single("License")
	output.add_single("Keywords")

	# URLs
	output.add_single("Home-page")
	output.add_single("Download-URL")
	output.add_multiple("Project-URL")

	# Platforms
	output.add_multiple("Platform")
	output.add_multiple("Supported-Platform")
	output.add_multiple("Classifier")

	# Requirements
	output.add_single("Requires-Python")
	output.add_multiple("Requires-Dist")
	output.add_multiple("Provides-Extra")
	output.add_multiple("Requires-External")
	output.add_multiple("Provides-Dist")
	output.add_multiple("Obsoletes-Dist")

	# Description
	output.add_single("Description-Content-Type")

	if "Description" in fields:
		output.add_body(fields["Description"])

	return str(output)


def dump(fields: MetadataMapping, filename: PathLike) -> int:
	"""
	Construct Python core metadata from the given fields, and write it to ``filename``.

	:param fields:
	:param filename:
	"""

	filename = PathPlus(filename)
	return filename.write_text(dumps(fields))



#!/usr/bin/env python3
#
#  wheel.py
"""
Parse and create ``*dist-info/WHEEL`` files.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Any, List, Mapping, Union

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from domdf_python_tools.utils import divide, strtobool

# this package
from dist_meta.metadata import MetadataEmitter, MissingFieldError
from dist_meta.metadata_mapping import MetadataMapping

__all__ = ["dump", "dumps", "load", "loads"]


def loads(rawtext: str) -> MetadataMapping:
	"""
	Parse a ``WHEEL`` file from the given string.

	:param rawtext:

	:returns: A mapping of the metadata fields, and the long description
	"""

	file_content: List[str] = rawtext.splitlines()
	file_content.reverse()

	fields: MetadataMapping = MetadataMapping()

	while file_content:
		line = file_content.pop()
		if line.strip():
			field_name, field_value = divide(line, ':')
			fields[field_name] = field_value.lstrip()

	if "Wheel-Version" not in fields:
		raise MissingFieldError(f"No 'Wheel-Version' field was provided.")

	return fields


def load(filename: PathLike) -> MetadataMapping:
	"""
	Parse a ``WHEEL`` file from the given file.

	:param filename:

	:returns: A mapping of the metadata fields, and the long description
	"""

	filename = PathPlus(filename)
	return loads(filename.read_text())


def dumps(fields: Union[Mapping[str, Any], MetadataMapping]) -> str:
	"""
	Construct a ``WHEEL`` file from the given fields.

	:param fields: May be a conventional mapping, with ``Root-Is-Purelib`` as a boolean
		and ``Tag`` as a list of strings.
	"""

	output = MetadataEmitter(fields)  # type: ignore

	if "Wheel-Version" in fields:
		output.append(f"Wheel-Version: {float(fields['Wheel-Version'])}")
	else:
		raise MissingFieldError(f"No 'Wheel-Version' field was provided.")

	output.add_single("Generator")

	root_is_purelib = strtobool(fields.get("Root-Is-Purelib", False))
	output.append(f"Root-Is-Purelib: {str(root_is_purelib).lower()}")

	if "Tag" in fields and isinstance(fields, MetadataMapping):
		output.add_multiple("Tag")
	elif "Tag" in fields:
		for value in fields["Tag"]:
			output.append(f"Tag: {value}")

	output.add_single("Build")

	return str(output)


def dump(fields: Union[Mapping[str, Any], MetadataMapping], filename: PathLike) -> int:
	"""
	Construct a ``WHEEL`` file from the given fields, and write it to ``filename``.

	:param fields: May be a conventional mapping, with ``Root-Is-Purelib`` as a boolean
		and ``Tag`` as a list of strings.
	:param filename:
	"""

	filename = PathPlus(filename)
	return filename.write_text(dumps(fields))



#!/usr/bin/env python
#
#  metadata_mapping.py
"""
:class:`collections.abc.MutableMapping` which supports duplicate, case-insensitive keys.

.. caution::

	These are pretty low-level classes. You probably don't need to use these directly
	unless you're customising the ``METADATA`` file creation or parsing.

"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Based on CPython.
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2021 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#

# stdlib
from typing import Iterator, List, MutableMapping, Optional, Tuple, TypeVar, Union, cast, overload

# 3rd party
from domdf_python_tools.stringlist import DelimitedList, StringList

__all__ = ["MetadataMapping", "MetadataEmitter"]

_T = TypeVar("_T")


class MetadataMapping(MutableMapping[str, str]):
	"""
	Provides a :class:`~collections.abc.MutableMapping` interface to a list of fields,
	such as those used for Python `core metadata`_.

	.. _core metadata: https://packaging.python.org/specifications/core-metadata/

	.. seealso:: :class:`email.message.Message` and :class:`email.message.EmailMessage`

	Implements the :class:`~collections.abc.MutableMapping` interface,
	which assumes there is exactly one occurrence of the field per mapping.
	Some fields do in fact appear multiple times,
	and for those fields you must use the :meth:`~.MetadataMapping.get_all` method
	to obtain all values for that field.
	"""  # noqa: D400

	def __init__(self):
		self._fields: List[Tuple[str, str]] = []

	#
	# MAPPING INTERFACE (partial)
	#

	def __len__(self) -> int:
		"""
		Return the total number of keys, including duplicates.
		"""

		return len(self._fields)

	def __getitem__(self, name: str) -> str:
		"""
		Get a field value.

		.. note::

			If the field appears multiple times, exactly which occurrence gets returned is undefined.
			Use the :meth:`~.MetadataMapping.get_all` method to get all values matching a field name.

		.. latex:vspace:: -10px

		:param name:
		"""

		if name not in self:
			raise KeyError(name)

		return cast(str, self.get(name))

	def __setitem__(self, name: str, val: str) -> None:
		"""
		Set the value of a field.

		.. attention:

			This does not overwrite existing fields with the same field name.
			Use :meth:`MetadataMapping.__delitem__` first to delete any existing fields.

		:param name:
		:param val:
		"""

		self._fields.append((name, val))

	def __delitem__(self, name: str) -> None:
		"""
		Delete all occurrences of a field, if present.

		Does not raise an exception if the field is missing.

		:param name:
		"""

		name = name.lower()
		newfields = []

		for k, v in self._fields:
			if k.lower() != name:
				newfields.append((k, v))

		self._fields = newfields

	def __contains__(self, name: object) -> bool:
		"""
		Returns whether ``name`` is in the :class:`~.MetadataMapping`.

		:param name:
		"""

		if not isinstance(name, str):
			return False

		name = name.lower()

		for k, v in self._fields:
			if k.lower() == name:
				return True

		return False

	def __iter__(self) -> Iterator[str]:
		"""
		Returns an iterator over the keys in the :class:`~.MetadataMapping`.
		"""

		for field, value in self._fields:
			yield field

	def keys(self) -> List[str]:  # type: ignore
		"""
		Return a list of all field *names*.

		These will be sorted by insertion order, and may contain duplicates.
		Any fields deleted and re-inserted are always appended to the field list.
		"""

		return [k for k, v in self._fields]

	def values(self) -> List[str]:  # type: ignore
		"""
		Return a list of all field *values*.

		These will be sorted by insertion order, and may contain duplicates.
		Any fields deleted and re-inserted are always appended to the field list.
		"""

		return [v for k, v in self._fields]

	def items(self) -> List[Tuple[str, str]]:  # type: ignore
		"""
		Get all the fields and their values.

		These will be sorted by insertion order, and may contain duplicates.
		Any fields deleted and re-inserted are always appended to the field list.
		"""

		return self._fields[:]

	@overload
	def get(self, name: str) -> Optional[str]: ...

	@overload
	def get(self, name: str, default: Union[str, _T]) -> Union[str, _T]: ...

	def get(self, name: str, default=None):
		"""
		Get a field value.

		Like :meth:`~.MetadataMapping.__getitem__`,
		but returns ``default`` instead of :py:obj:`None` when the field is missing.

		.. note::

			If the field appears multiple times, exactly which occurrence gets returned is undefined.
			Use the :meth:`~.MetadataMapping.get_all` method to get all values matching a field name.

		.. latex:vspace:: -10px

		:param name:
		:param default:
		"""

		name = name.lower()

		for key, val in self._fields:
			if key.lower() == name:
				return val

		return default

	#
	# Additional useful stuff
	#

	@overload
	def get_all(self, name: str) -> Optional[List[str]]: ...

	@overload
	def get_all(self, name: str, default: Union[str, _T]) -> Union[List[str], _T]: ...

	def get_all(self, name: str, default=None):
		"""
		Return a list of all the values for the named field.

		These will be sorted in the order they appeared in the original message,
		and may contain duplicates.
		Any fields deleted and re-inserted are always appended to the field list.

		If no such fields exist, ``default`` is returned.

		:param name:
		:param default:
		"""

		values = []
		name = name.lower()

		for key, val in self._fields:
			if key.lower() == name:
				values.append(val)

		if not values:
			return default

		return values

	def __repr__(self) -> str:
		"""
		Return a string representation of the :class:`~.MetadataMapping`.
		"""

		items = DelimitedList([f"{k!r}: {v!r}" for k, v in self.items()])
		as_dict = f"{{{items:, }}}"

		return f"<{self.__class__.__name__}({as_dict})>"

	def replace(self, name: str, value: str):
		"""
		Replace the value of the first matching field, retaining header order and case.

		:raises KeyError: If no matching field was found.
		"""

		for i, (key, val) in enumerate(self._fields):
			if key.lower() == name.lower():
				self._fields[i] = (name, value)
				break
		else:
			raise KeyError(name)


class MetadataEmitter(StringList):
	"""
	Used to construct ``METADATA``, ``WHEEL`` and other email field-like files.

	:param fields: The fields the file is being constructed from.

	.. autosummary-widths:: 1/3
		:html: 3/10
	"""

	def __init__(self, fields: MetadataMapping):
		self.fields = fields
		super().__init__()

	def add_single(self, field_name: str):
		"""
		Add a single value for the field with the given name.

		:param field_name:
		"""

		if field_name in self.fields:
			self.append(f"{field_name}: {self.fields[field_name]}")

	def add_multiple(self, field_name: str):
		"""
		Add all values for the "multiple use" field with the given name.

		:param field_name:
		"""

		if field_name in self.fields:
			for value in self.fields.get_all(field_name, ()):
				self.append(f"{field_name}: {value}")

	def add_body(self, body: str):
		"""
		Add a body to the file.

		In an email message this is the message content itself.

		:param body:
		"""

		self.blankline(ensure_single=True)
		self.blankline()
		self.append(body)
		self.blankline(ensure_single=True)



#!/usr/bin/env python3
#
#  record.py
"""
Classes to model parts of ``RECORD`` files.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import csv
import os
import pathlib
from base64 import urlsafe_b64decode, urlsafe_b64encode
from typing import TYPE_CHECKING, NamedTuple, Optional, Type, TypeVar

# 3rd party
from domdf_python_tools.stringlist import DelimitedList
from domdf_python_tools.typing import PathLike

if TYPE_CHECKING:
	# stdlib
	from hashlib import _Hash

	# this package
	from dist_meta.distributions import Distribution
else:
	try:
		# 3rd party
		from _hashlib import HASH as _Hash
	except ImportError:  # pragma: no cover
		try:
			# 3rd party
			from _hashlib import Hash as _Hash
		except ImportError:
			pass

__all__ = ["FileHash", "RecordEntry"]

_RE = TypeVar("_RE", bound="RecordEntry")
_FH = TypeVar("_FH", bound="FileHash")


class RecordEntry(pathlib.PurePosixPath):
	"""
	Represents a path in a distribution.

	:param path: The path to the file in the distribution, relative to the distribution root
		(i.e. the ``site-packages`` directory).
	:param hash: The hash/checksum of the file.
	:param size: The size of the file.
	:param distro: The distribution the file belongs to.

	.. note::

		Path operations (:meth:`~.pathlib.PurePath.joinpath`, :data:`~.pathlib.PurePath.parent` etc.)
		will return a standard :class:`pathlib.PurePosixPath` object without the extended attributes of this class.

	.. latex:vspace:: -10px
	"""

	__slots__ = ("hash", "size", "distro")

	#: The name of the file in the distribution.
	name: str

	#: The hash/checksum of the file.
	hash: Optional["FileHash"]  # noqa: A003  # pylint: disable=redefined-builtin

	#: The size of the file.
	size: Optional[int]

	#: The distribution the file belongs to.
	distro: Optional["Distribution"]

	def __init__(
			self,
			path: PathLike,
			hash: Optional["FileHash"] = None,  # noqa: A002  # pylint: disable=redefined-builtin
			size: Optional[int] = None,
			distro: Optional["Distribution"] = None,
			):
		super().__init__()

	def __new__(
			cls: Type[_RE],
			path: PathLike,
			hash: Optional["FileHash"] = None,  # noqa: A002  # pylint: disable=redefined-builtin
			size: Optional[int] = None,
			distro: Optional["Distribution"] = None,
			) -> _RE:
		"""
		Construct a :class:`RecordEntry` from one a string or an existing :class:`pathlib.PurePath` object.
		"""

		self = cls._from_parts((path, ))  # type: ignore
		self.hash = hash
		self.size = size
		self.distro = distro
		return self

	def read_text(
			self,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			) -> str:
		"""
		Open the file in text mode, read it, and close the file.

		:param encoding: The encoding to write to the file in.
		:param errors:

		:return: The content of the file.

		.. attention:: This operation requires a value for :attr:`self.distro <.RecordEntry.distro>`.
		"""

		if self.distro is None:
			raise ValueError("Cannot read files with 'self.distro = None'")

		return (self.distro.path.parent / self).read_text(encoding=encoding, errors=errors)

	def read_bytes(self):
		"""
		Open the file in bytes mode, read it, and close the file.

		:return: The content of the file.

		.. attention:: This operation requires a value for :attr:`self.distro <.RecordEntry.distro>`.
		"""

		if self.distro is None:
			raise ValueError("Cannot read files with 'self.distro = None'")

		return (self.distro.path.parent / self).read_bytes()

	def __repr__(self) -> str:
		"""
		Return a string representation of the :class:`~.RecordEntry`.
		"""

		parts = DelimitedList([f"{os.fspath(self)!r}"])
		if self.hash is not None:
			parts.append(f"hash={self.hash!r}")
		if self.size is not None:
			parts.append(f"size={self.size!r}")
		if self.distro is not None:
			parts.append(f"distro={self.distro!r}")
		return f"{self.__class__.__name__}({parts:, })"

	def as_record_entry(self):
		"""
		Returns an entry for a ``RECORD`` file, in the form ``<name>,<hash>,<size>``.
		"""

		parts = [os.fspath(self)]

		if self.hash is not None:
			parts.append(self.hash.to_string())
		else:
			parts.append('')

		if self.size is not None:
			parts.append(str(self.size))
		else:
			parts.append('')

		return ','.join(parts)

	@classmethod
	def from_record_entry(
			cls: Type[_RE],
			entry: str,
			distro: Optional["Distribution"] = None,
			) -> _RE:
		"""
		Construct a :class:`~.RecordEntry` from a line in a ``RECORD`` file, in the form ``<name>,<hash>,<size>``.

		.. versionadded:: 0.2.0

		:param entry:
		:param distro: The distribution the ``RECORD`` file belongs to. Optional.

		:rtype: :class:`~.RecordEntry`
		"""

		entry = entry.strip()
		lines = entry.splitlines()

		if len(lines) != 1:
			raise ValueError("'entry' must be a single-line entry.")

		entry = lines[0]

		if '"' in entry:
			name, hash_, size_str, *_ = next(csv.reader((entry, )))
		else:
			name, hash_, size_str, *_ = entry.split(',')

		hash_ = hash_.strip()
		size_str = size_str.strip()

		return cls(
				name.strip(),
				hash=FileHash.from_string(hash_) if hash_ else None,
				size=int(size_str) if size_str else None,
				distro=distro,
				)


class FileHash(NamedTuple):
	"""
	Represents a checksum for a file in a ``RECORD`` file, or as the URL fragment in a :pep:`503` repository URL.
	"""

	#: The name of the hash algorithm.
	name: str

	#: The :func:`~.base64.urlsafe_b64encode`'d hexdigest of the hash.
	value: str

	@classmethod
	def from_string(cls: Type[_FH], string: str) -> _FH:
		"""
		Constructs a  :class:`~.FileHash` from a  string in the form ``<name>=<value>``.

		:param string:

		:rtype: :class:`~.FileHash`
		"""

		name, _, value = string.partition('=')
		return cls(name.strip(), value.strip())

	def to_string(self) -> str:
		"""
		Returns the :class:`~.FileHash` as a string, in the form ``<name>=<value>``.
		"""

		return f"{self.name}={self.value}"

	def digest(self) -> bytes:
		"""
		Returns the digest of the hash.

		This is a bytes object which may contain bytes in the whole range from 0 to 255.
		"""

		return urlsafe_b64decode(f"{self.value}==".encode("latin1"))

	def hexdigest(self) -> str:
		"""
		Like :meth:`self.digest() <.FileHash.digest>` except the digest is returned as
		a string object of double length, containing only hexadecimal digits.

		This may be used to exchange the value safely in email or other non-binary environments.
		"""  # noqa: D400

		return ''.join(f"{x:0{2}x}" for x in self.digest())

	@classmethod
	def from_hash(cls: Type[_FH], the_hash: "_Hash") -> _FH:
		"""
		Construct a :class:`~.FileHash` object from a :mod:`hashlib` hash object.

		:param the_hash:
		:type the_hash: :mod:`hashlib.HASH <hashlib>`

		:rtype: :class:`~.FileHash`
		"""

		name = the_hash.name
		value = urlsafe_b64encode(the_hash.digest()).decode("latin1").rstrip('=')
		return cls(name, value)



#!/usr/bin/env python3
#
#  _utils.py
"""
Private utility functions.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import functools
import os
import pathlib
import re
from typing import Iterator, Tuple

# 3rd party
from domdf_python_tools.paths import PathPlus
from packaging.utils import InvalidWheelFilename, canonicalize_name
from packaging.version import Version

_escaped_name_re = re.compile(r"^[\w\d._]*$", re.UNICODE)

SHOULD_CACHE = int(os.environ.get("DIST_META_CACHE", 1))


def _cache(func):
	if SHOULD_CACHE:
		return functools.lru_cache()(func)
	else:  # pragma: no cover
		return func


@_cache
def _canonicalize(name: str) -> str:
	return canonicalize_name(name)


@_cache
def _parse_wheel_filename(filename: pathlib.PurePath) -> Tuple[str, Version]:
	# From https://github.com/pypa/packaging
	# This software is made available under the terms of *either* of the licenses
	# found in LICENSE.APACHE or LICENSE.BSD. Contributions to this software is made
	# under the terms of *both* these licenses.

	if not filename.suffix == ".whl":
		raise InvalidWheelFilename(f"Invalid wheel filename (extension must be '.whl'): {filename}")

	stem = filename.stem
	dashes = stem.count('-')
	if dashes not in (4, 5):
		raise InvalidWheelFilename(f"Invalid wheel filename (wrong number of parts): {filename}")

	parts = stem.split('-', dashes - 2)
	name = parts[0]
	# See PEP 427 for the rules on escaping the project name
	if "__" in name or _escaped_name_re.match(name) is None:
		raise InvalidWheelFilename(f"Invalid project name: {name!r}")

	version = _parse_version(parts[1])

	return name, version


def _iter_dist_infos(basedir: PathPlus) -> Iterator[os.DirEntry]:
	subdir: os.DirEntry
	with os.scandir(os.fspath(basedir)) as sd:
		for subdir in sd:
			if not subdir.is_dir():
				continue

			if not subdir.name.endswith(".dist-info"):
				continue

			yield subdir


@_cache
def _parse_version(version: str) -> Version:
	return Version(version)
